"""
Python 2 & 3 Compatibility Functions

"""


def ascii(value):
    """Return the string of value

    :param mixed value: The value to return
    :rtype: str

    """
    return '{0}'.format(value)



"""Redis Scripting Commands Mixin"""


class ScriptingMixin(object):
    """Redis Scripting Commands Mixin"""

    def eval(self, script, keys=None, args=None):
        """:meth:`~tredis.RedisClient.eval` and
        :meth:`~tredis.RedisClient.evalsha` are used to evaluate scripts using
        the Lua interpreter built into Redis starting from version 2.6.0.

        The first argument of EVAL is a Lua 5.1 script. The script does not
        need to define a Lua function (and should not). It is just a Lua
        program that will run in the context of the Redis server.

        .. note::

           **Time complexity**: Depends on the script that is executed.

        :param str script: The Lua script to execute
        :param list keys: A list of keys to pass into the script
        :param list args: A list of args to pass into the script
        :return: mixed

        """
        if not keys:
            keys = []
        if not args:
            args = []
        return self._execute([b'EVAL', script, str(len(keys))] + keys + args)

    def evalsha(self, sha1, keys=None, args=None):
        """Evaluates a script cached on the server side by its SHA1 digest.
        Scripts are cached on the server side using the
        :meth:`~tredis.RedisClient.script_load` command. The command is
        otherwise identical to :meth:`~tredis.RedisClient.eval`.

        .. note::

           **Time complexity**: Depends on the script that is executed.

        :param str sha1: The sha1 hash of the script to execute
        :param list keys: A list of keys to pass into the script
        :param list args: A list of args to pass into the script
        :return: mixed

        """
        if not keys:
            keys = []
        if not args:
            args = []
        return self._execute([b'EVALSHA', sha1, str(len(keys))] + keys + args)

    def script_exists(self, *hashes):
        """Returns information about the existence of the scripts in the script
        cache.

        This command accepts one or more SHA1 digests and returns a list of
        ones or zeros to signal if the scripts are already defined or not
        inside the script cache. This can be useful before a pipelining
        operation to ensure that scripts are loaded (and if not, to load them
        using :meth:`~tredis.RedisClient.script_load`) so that the pipelining
        operation can be performed solely using
        :meth:`~tredis.RedisClient.evalsha` instead of
        :meth:`~tredis.RedisClient.eval` to save bandwidth.

        Please refer to the :meth:`~tredis.RedisClient.eval` documentation for
        detailed information about Redis Lua scripting.

        .. note::

           **Time complexity**: ``O(N)`` with ``N`` being the number of scripts
           to check (so checking a single script is an ``O(1)`` operation).

        :param str hashes: One or more sha1 hashes to check for in the cache
        :rtype: list
        :return: Returns a list of ``1`` or ``0`` indicating if the specified
            script(s) exist in the cache.

        """
        return self._execute([b'SCRIPT', b'EXISTS'] + list(hashes))

    def script_flush(self):
        """Flush the Lua scripts cache.

        Please refer to the :meth:`~tredis.RedisClient.eval` documentation for
        detailed information about Redis Lua scripting.

        .. note::

           **Time complexity**: ``O(N)`` with ``N`` being the number of scripts
           in cache

        :rtype: bool

        """
        return self._execute([b'SCRIPT', b'FLUSH'], b'OK')

    def script_kill(self):
        """Kills the currently executing Lua script, assuming no write
        operation was yet performed by the script.

        This command is mainly useful to kill a script that is running for too
        much time(for instance because it entered an infinite loop because of
        a bug). The script will be killed and the client currently blocked into
        :meth:`~tredis.RedisClient.eval` will see the command returning with an
        error.

        If the script already performed write operations it can not be killed
        in this way because it would violate Lua script atomicity contract. In
        such a case only SHUTDOWN NOSAVE is able to kill the script, killing
        the Redis process in an hard way preventing it to persist with
        half-written information.

        Please refer to the :meth:`~tredis.RedisClient.eval` documentation for
        detailed information about Redis Lua scripting.

        .. note::

           **Time complexity**: ``O(1)``

        :rtype: bool

        """
        return self._execute([b'SCRIPT', b'KILL'], b'OK')

    def script_load(self, script):
        """Load a script into the scripts cache, without executing it. After
        the specified command is loaded into the script cache it will be
        callable using :meth:`~tredis.RedisClient.evalsha` with the correct
        SHA1 digest of the script, exactly like after the first successful
        invocation of :meth:`~tredis.RedisClient.eval`.

        The script is guaranteed to stay in the script cache forever (unless
        :meth:`~tredis.RedisClient.script_flush` is called).

        The command works in the same way even if the script was already
        present in the script cache.

        Please refer to the :meth:`~tredis.RedisClient.eval` documentation for
        detailed information about Redis Lua scripting.

        .. note::

           **Time complexity**: ``O(N)`` with ``N`` being the length in bytes
           of the script body.

        :param str script: The script to load into the script cache
        :return: str

        """
        return self._execute([b'SCRIPT', b'LOAD', script])



"""Redis Set Commands Mixin"""
from tornado import concurrent

# Python 2 support for ascii()
if 'ascii' not in dir(__builtins__):  # pragma: nocover
    from tredis.compat import ascii


class SetsMixin(object):
    """Redis Set Commands Mixin"""

    def sadd(self, key, *members):
        """Add the specified members to the set stored at key. Specified
        members that are already a member of this set are ignored. If key does
        not exist, a new set is created before adding the specified members.

        An error is returned when the value stored at key is not a set.

        Returns :data:`True` if all requested members are added. If more
        than one member is passed in and not all members are added, the
        number of added members is returned.

        .. note::

           **Time complexity**: ``O(N)`` where ``N`` is the number of members
           to be added.

        :param key: The key of the set
        :type key: :class:`str`, :class:`bytes`
        :param members: One or more positional arguments to add to the set
        :type key: :class:`str`, :class:`bytes`
        :returns: Number of items added to the set
        :rtype: bool, int

        """
        return self._execute([b'SADD', key] + list(members), len(members))

    def scard(self, key):
        """Returns the set cardinality (number of elements) of the set stored
        at key.

        .. note::

           **Time complexity**: ``O(1)``

        :param key: The key of the set
        :type key: :class:`str`, :class:`bytes`
        :rtype: int
        :raises: :exc:`~tredis.exceptions.RedisError`

        """
        return self._execute([b'SCARD', key])

    def sdiff(self, *keys):
        """Returns the members of the set resulting from the difference between
        the first set and all the successive sets.

        For example:

        .. code::

            key1 = {a,b,c,d}
            key2 = {c}
            key3 = {a,c,e}
            SDIFF key1 key2 key3 = {b,d}

        Keys that do not exist are considered to be empty sets.

        .. note::

           **Time complexity**: ``O(N)`` where ``N`` is the total number of
           elements in all given sets.

        :param keys: Two or more set keys as positional arguments
        :type keys: :class:`str`, :class:`bytes`
        :rtype: list
        :raises: :exc:`~tredis.exceptions.RedisError`

        """
        return self._execute([b'SDIFF'] + list(keys))

    def sdiffstore(self, destination, *keys):
        """This command is equal to :meth:`~tredis.RedisClient.sdiff`, but
        instead of returning the resulting set, it is stored in destination.

        If destination already exists, it is overwritten.

        .. note::

           **Time complexity**: ``O(N)`` where ``N`` is the total number of
           elements in all given sets.

        :param destination: The set to store the diff into
        :type destination: :class:`str`, :class:`bytes`
        :param keys: One or more set keys as positional arguments
        :type keys: :class:`str`, :class:`bytes`
        :rtype: int
        :raises: :exc:`~tredis.exceptions.RedisError`

        """
        return self._execute([b'SDIFFSTORE', destination] + list(keys))

    def sinter(self, *keys):
        """Returns the members of the set resulting from the intersection of
        all the given sets.

        For example:

        .. code::

            key1 = {a,b,c,d}
            key2 = {c}
            key3 = {a,c,e}
            SINTER key1 key2 key3 = {c}

        Keys that do not exist are considered to be empty sets. With one of
        the keys being an empty set, the resulting set is also empty (since
        set intersection with an empty set always results in an empty set).

        .. note::

           **Time complexity**: ``O(N*M)`` worst case where ``N`` is the
           cardinality of the smallest set and ``M`` is the number of sets.

        :param keys: Two or more set keys as positional arguments
        :type keys: :class:`str`, :class:`bytes`
        :rtype: list
        :raises: :exc:`~tredis.exceptions.RedisError`

        """
        return self._execute([b'SINTER'] + list(keys))

    def sinterstore(self, destination, *keys):
        """This command is equal to :meth:`~tredis.RedisClient.sinter`, but
        instead of returning the resulting set, it is stored in destination.

        If destination already exists, it is overwritten.

        .. note::

           **Time complexity**: ``O(N*M)`` worst case where ``N`` is the
           cardinality of the smallest set and ``M`` is the number of sets.

        :param destination: The set to store the intersection into
        :type destination: :class:`str`, :class:`bytes`
        :param keys: One or more set keys as positional arguments
        :type keys: :class:`str`, :class:`bytes`
        :rtype: int
        :raises: :exc:`~tredis.exceptions.RedisError`

        """
        return self._execute([b'SINTERSTORE', destination] + list(keys))

    def sismember(self, key, member):
        """Returns :data:`True` if ``member`` is a member of the set stored
        at key.

        .. note::

           **Time complexity**: ``O(1)``

        :param key: The key of the set to check for membership in
        :type key: :class:`str`, :class:`bytes`
        :param member: The value to check for set membership with
        :type member: :class:`str`, :class:`bytes`
        :rtype: bool
        :raises: :exc:`~tredis.exceptions.RedisError`

        """
        return self._execute([b'SISMEMBER', key, member], 1)

    def smembers(self, key):
        """Returns all the members of the set value stored at key.

        This has the same effect as running :meth:`~tredis.RedisClient.sinter`
        with one argument key.

        .. note::

           **Time complexity**: ``O(N)`` where ``N`` is the set cardinality.

        :param key: The key of the set to return the members from
        :type key: :class:`str`, :class:`bytes`
        :rtype: list
        :raises: :exc:`~tredis.exceptions.RedisError`

        """
        return self._execute([b'SMEMBERS', key])

    def smove(self, source, destination, member):
        """Move member from the set at source to the set at destination. This
        operation is atomic. In every given moment the element will appear to
        be a member of source or destination for other clients.

        If the source set does not exist or does not contain the specified
        element, no operation is performed and :data:`False` is returned.
        Otherwise, the element is removed from the source set and added to the
        destination set. When the specified element already exists in the
        destination set, it is only removed from the source set.

        An error is returned if source or destination does not hold a set
        value.

        .. note::

           **Time complexity**: ``O(1)``

        :param source: The source set key
        :type source: :class:`str`, :class:`bytes`
        :param destination: The destination set key
        :type destination: :class:`str`, :class:`bytes`
        :param member: The member value to move
        :type member: :class:`str`, :class:`bytes`
        :rtype: bool
        :raises: :exc:`~tredis.exceptions.RedisError`

        """
        return self._execute([b'SMOVE', source, destination, member], 1)

    def spop(self, key, count=None):
        """Removes and returns one or more random elements from the set value
        store at key.

        This operation is similar to :meth:`~tredis.RedisClient.srandmember`,
        that returns one or more random elements from a set but does not remove
        it.

        The count argument will be available in a later version and is not
        available in 2.6, 2.8, 3.0

        Redis 3.2 will be the first version where an optional count argument
        can be passed to :meth:`~tredis.RedisClient.spop` in order to retrieve
        multiple elements in a single call. The implementation is already
        available in the unstable branch.

        .. note::

           **Time complexity**: Without the count argument ``O(1)``, otherwise
           ``O(N)`` where ``N`` is the absolute value of the passed count.

        :param key: The key to get one or more random members from
        :type key: :class:`str`, :class:`bytes`
        :param int count: The number of members to return
        :rtype: bytes, list
        :raises: :exc:`~tredis.exceptions.RedisError`

        """
        command = [b'SPOP', key]
        if count:  # pragma: nocover
            command.append(ascii(count).encode('ascii'))
        return self._execute(command)

    def srandmember(self, key, count=None):
        """When called with just the key argument, return a random element from
        the set value stored at key.

        Starting from Redis version 2.6, when called with the additional count
        argument, return an array of count distinct elements if count is
        positive. If called with a negative count the behavior changes and the
        command is allowed to return the same element multiple times. In this
        case the number of returned elements is the absolute value of the
        specified count.

        When called with just the key argument, the operation is similar to
        :meth:`~tredis.RedisClient.spop`, however while
        :meth:`~tredis.RedisClient.spop` also removes the randomly selected
        element from the set, :meth:`~tredis.RedisClient.srandmember` will just
        return a random element without altering the original set in any way.

        .. note::

           **Time complexity**: Without the count argument ``O(1)``, otherwise
           ``O(N)`` where ``N`` is the absolute value of the passed count.

        :param key: The key to get one or more random members from
        :type key: :class:`str`, :class:`bytes`
        :param int count: The number of members to return
        :rtype: bytes, list
        :raises: :exc:`~tredis.exceptions.RedisError`

        """
        command = [b'SRANDMEMBER', key]
        if count:
            command.append(ascii(count).encode('ascii'))
        return self._execute(command)

    def srem(self, key, *members):
        """Remove the specified members from the set stored at key. Specified
        members that are not a member of this set are ignored. If key does not
        exist, it is treated as an empty set and this command returns ``0``.

        An error is returned when the value stored at key is not a set.

        Returns :data:`True` if all requested members are removed. If more
        than one member is passed in and not all members are removed, the
        number of removed members is returned.

        .. note::

           **Time complexity**: ``O(N)`` where ``N`` is the number of members
           to be removed.

        :param key: The key to remove the member from
        :type key: :class:`str`, :class:`bytes`
        :param mixed members: One or more member values to remove
        :rtype: bool, int
        :raises: :exc:`~tredis.exceptions.RedisError`

        """
        return self._execute([b'SREM', key] + list(members), len(members))

    def sscan(self, key, cursor=0, pattern=None, count=None):
        """The :meth:`~tredis.RedisClient.sscan` command and the closely
        related commands :meth:`~tredis.RedisClient.scan`,
        :meth:`~tredis.RedisClient.hscan` and :meth:`~tredis.RedisClient.zscan`
        are used in order to incrementally iterate over a collection of
        elements.

        - :meth:`~tredis.RedisClient.scan` iterates the set of keys in the
          currently selected Redis database.
        - :meth:`~tredis.RedisClient.sscan` iterates elements of Sets types.
        - :meth:`~tredis.RedisClient.hscan` iterates fields of Hash types and
          their associated values.
        - :meth:`~tredis.RedisClient.zscan` iterates elements of Sorted Set
          types and their associated scores.

        **Basic usage**

        :meth:`~tredis.RedisClient.sscan` is a cursor based iterator. This
        means that at every call of the command, the server returns an updated
        cursor that the user needs to use as the cursor argument in the next
        call.

        An iteration starts when the cursor is set to ``0``, and terminates
        when the cursor returned by the server is ``0``.

        For more information on :meth:`~tredis.RedisClient.scan`,
        visit the `Redis docs on scan <http://redis.io/commands/scan>`_.

        .. note::

           **Time complexity**: ``O(1)`` for every call. ``O(N)`` for a
           complete iteration, including enough command calls for the cursor to
           return back to ``0``. ``N`` is the number of elements inside the
           collection.

        :param key: The key to scan
        :type key: :class:`str`, :class:`bytes`
        :param int cursor: The server specified cursor value or ``0``
        :param pattern: An optional pattern to apply for key matching
        :type pattern: :class:`str`, :class:`bytes`
        :param int count: An optional amount of work to perform in the scan
        :rtype: int, list
        :returns: A tuple containing the cursor and the list of set items
        :raises: :exc:`~tredis.exceptions.RedisError`

        """

        def format_response(value):
            """Format the response from redis

            :param tuple value: The return response from redis
            :rtype: tuple(int, list)

            """
            return int(value[0]), value[1]

        command = [b'SSCAN', key, ascii(cursor).encode('ascii')]
        if pattern:
            command += [b'MATCH', pattern]
        if count:
            command += [b'COUNT', ascii(count).encode('ascii')]
        return self._execute(command, format_callback=format_response)

    def sunion(self, *keys):
        """Returns the members of the set resulting from the union of all the
        given sets.

        For example:

        .. code::

            key1 = {a,b,c,d}
            key2 = {c}
            key3 = {a,c,e}
            SUNION key1 key2 key3 = {a,b,c,d,e}

        .. note::

           **Time complexity**: ``O(N)`` where ``N`` is the total number of
           elements in all given sets.

        Keys that do not exist are considered to be empty sets.

        :param keys: Two or more set keys as positional arguments
        :type keys: :class:`str`, :class:`bytes`
        :rtype: list
        :raises: :exc:`~tredis.exceptions.RedisError`

        """
        return self._execute([b'SUNION'] + list(keys))

    def sunionstore(self, destination, *keys):
        """This command is equal to :meth:`~tredis.RedisClient.sunion`, but
        instead of returning the resulting set, it is stored in destination.

        If destination already exists, it is overwritten.

        .. note::

           **Time complexity**: ``O(N)`` where ``N`` is the total number of
           elements in all given sets.

        :param destination: The set to store the union into
        :type destination: :class:`str`, :class:`bytes`
        :param keys: One or more set keys as positional arguments
        :type keys: :class:`str`, :class:`bytes`
        :rtype: int
        :raises: :exc:`~tredis.exceptions.RedisError`

        """
        return self._execute([b'SUNIONSTORE', destination] + list(keys))



"""
Cluster Supporting Redis Client

"""
import collections
import logging

import hiredis
from tornado import concurrent
from tornado import ioloop
from tornado import locks
from tornado import iostream
from tornado import tcpclient

from tredis import common
from tredis import crc16
from tredis import exceptions
from tredis import cluster
from tredis import connection
from tredis import geo
from tredis import hashes
from tredis import hyperloglog
from tredis import keys
from tredis import lists
from tredis import pubsub
from tredis import scripting
from tredis import server
from tredis import sets
from tredis import sortedsets
from tredis import strings
from tredis import transactions

LOGGER = logging.getLogger(__name__)

CRLF = b'\r\n'

DEFAULT_HOST = 'localhost'
"""The default host to connect to"""

DEFAULT_PORT = 6379
"""The default port to connect to"""

DEFAULT_DB = 0
"""The default database number to use"""

HASH_SLOTS = 16384
"""Redis Cluster Hash Slots Value"""

# Python 2 support for ascii()
if 'ascii' not in dir(__builtins__):  # pragma: nocover
    from tredis.compat import ascii


Command = collections.namedtuple('Command', ['command', 'connection',
                                             'expectation', 'callback'])


class _Connection(object):
    """Manages the redis TCP connection.

    :param str host: The hostname to connect to
    :param int port: The port to connect on
    :param int db: The database number to use
    :param method on_close: The method to call if the connection is closed

    """

    def __init__(self, host, port, db, on_written, on_close, io_loop,
                 cluster_node=False, read_only=False, slots=None):
        super(_Connection, self).__init__()
        self.connected = False
        self.io_loop = io_loop
        self.host = host
        self.port = port
        self.database = int(db or DEFAULT_DB)

        self._client = tcpclient.TCPClient()
        self._cluster_node = cluster_node
        self._read_only = read_only
        self._slots = slots or []
        self._stream = None
        self._on_connect = None
        self._on_close = on_close
        self._on_written = on_written

    def close(self):
        """Close the stream.

        :raises: :class:`tredis.exceptions.ConnectionError` if the
            stream is not currently connected

        """
        if self._stream is None:
            raise exceptions.ConnectionError('Not connected')
        self._stream.close()

    def connect(self):
        """Connect to the Redis server if necessary.

        :rtype: :class:`~tornado.concurrent.Future`
        :raises: :class:`~tredis.exceptions.ConnectError`
                 :class:`~tredis.exceptinos.RedisError`

        """
        future = concurrent.Future()

        if self.connected:
            raise exceptions.ConnectError('already connected')

        LOGGER.debug('%s connecting', self.name)
        self.io_loop.add_future(self._client.connect(self.host, self.port),
                                lambda f: self._on_connected(f, future))
        return future

    def execute(self, command, future):
        """Execute a command after connecting if necessary.

        :param bytes command: command to execute after the connection
            is established
        :param tornado.concurrent.Future future:  future to resolve
            when the command's response is received.

        """
        LOGGER.debug('execute(%r, %r)', command, future)
        if self.connected:
            self._write(command, future)
        else:
            def on_connected(cfuture):
                if cfuture.exception():
                    return future.set_exception(cfuture.exception())
                self._write(command, future)

            self.io_loop.add_future(self.connect(), on_connected)

    @property
    def name(self):
        """Return the connection name as it is returned in the cluster nodes
        command.

        :rtype: str

        """
        return '{}:{}'.format(self.host, self.port)

    def read(self, callback):
        """Issue a read on the stream, invoke callback when completed.

        :raises: :class:`tredis.exceptions.ConnectionError` if the
            stream is not currently connected

        """
        self._stream.read_bytes(65536, callback, None, True)

    def set_read_only(self, read_only):
        """Change the connection's read-only flag in the client.

        :param bool read_only: Value to set
        """
        self._read_only = read_only

    def set_slots(self, slots):
        """Change the connection's slot list in the client.

        :param list slots: The updated slot values

        """
        self._slots = slots

    @property
    def slots(self):
        """Return the connection's slot values for clustering.

        :rtype: list

        """
        return self._slots

    def _on_closed(self):
        """Invoked when the connection is closed"""
        LOGGER.error('Redis connection closed')
        self.connected = False
        self._on_close()
        self._stream = None

    def _on_connected(self, stream_future, connect_future):
        """Invoked when the socket stream has connected, setting up the
        stream callbacks and invoking the on connect callback if set.

        :param stream_future: The connection socket future
        :type stream_future: :class:`~tornado.concurrent.Future`
        :param stream_future: The connection response future
        :type stream_future: :class:`~tornado.concurrent.Future`
        :raises: :exc:`tredis.exceptions.ConnectError`

        """
        if stream_future.exception():
            connect_future.set_exception(
                exceptions.ConnectError(stream_future.exception()))
        else:
            self._stream = stream_future.result()
            self._stream.set_close_callback(self._on_closed)
            self.connected = True
            connect_future.set_result(self)

    def _write(self, command, future):
        """Write a command to the socket

        :param Command command: the Command data structure

        """

        def on_written():
            self._on_written(command, future)

        try:
            self._stream.write(command.command, callback=on_written)
        except iostream.StreamClosedError as error:
            future.set_exception(exceptions.ConnectionError(error))
        except Exception as error:
            LOGGER.exception('unhandled write failure - %r', error)
            future.set_exception(exceptions.ConnectionError(error))


class Client(server.ServerMixin,
             keys.KeysMixin,
             strings.StringsMixin,
             geo.GeoMixin,
             hashes.HashesMixin,
             hyperloglog.HyperLogLogMixin,
             lists.ListsMixin,
             sets.SetsMixin,
             sortedsets.SortedSetsMixin,
             pubsub.PubSubMixin,
             connection.ConnectionMixin,
             cluster.ClusterMixin,
             scripting.ScriptingMixin,
             transactions.TransactionsMixin):
    """Asynchronous Redis client that supports Redis with master/slave failover
    and clustering. When ``clustering`` is ``True``, the client will
    automatically discover all of the nodes in the cluster and connect to them.

    The ``hosts`` argument should contain a list of Redis servers to connect
    to. The connection information for the server should be a :class:`dict`. In
    the following example, the client will connect to Redis running at
    ``127.0.0.1`` on port ``6379`` using database # ``2``:

    .. code:: python

        class RequestHandler(web.RequestHandler):

            @gen.coroutine
            def connect_to_redis(self)
                client = tredis.Client([{
                        'host': '127.0.0.1', 'port': 6379, 'db': 2
                    }, auto_connect=False, cluster=True]
                yield client.connect()

    When ``auto_connect`` is set to ``True``, the connection to the Redis
    server or the Redis cluster starts on creation of the client. You should be
    aware that this will not block on creation and the connection will be
    established asynchronously in the background. Any requests made with the
    client while it is connecting will block until the connection is available.

    When ``auto_connect`` is set to ``False``, you will need to invoke the
    :meth:`~tredis.Client.connect` method, yielding to the
    :class:`~tornado.concurrent.Future` that it returns.

    .. added: 0.7.0

    :param hosts: A list of host connection values.
    :type hosts: list(dict)
    :param io_loop: Override the current Tornado IOLoop instance
    :type io_loop: tornado.ioloop.IOLoop
    :param method on_close: The method to call if the connection is closed
    :param bool clustering: Toggle the cluster support in the client
    :param bool auto_connect: Toggle the auto-connect on creation feature


    """
    def __init__(self, hosts, on_close=None, io_loop=None, clustering=False,
                 auto_connect=True):
        """Create a new instance of the ``Client`` class.

        :param hosts: A list of host connection values.
        :type hosts: list(dict)
        :param io_loop: Override the current Tornado IOLoop instance
        :type io_loop: tornado.ioloop.IOLoop
        :param method on_close: The method to call if the connection is closed
        :param bool clustering: Toggle the cluster support in the client
        :param bool auto_connect: Toggle the auto-connect on creation feature

        """
        self._buffer = bytes()
        self._busy = locks.Lock()
        self._closing = False
        self._cluster = {}
        self._clustering = clustering
        self._connected = locks.Event()
        self._connect_future = concurrent.Future()
        self._connection = None
        self._discovery = False
        self._hosts = hosts
        self._on_close_callback = on_close
        self._reader = hiredis.Reader()
        self.io_loop = io_loop or ioloop.IOLoop.current()
        if not self._clustering:
            if len(hosts) > 1:
                raise ValueError('Too many hosts for non-clustering mode')
        if auto_connect:
            LOGGER.debug('Auto-connecting')
            self.connect()

    def connect(self):
        """Connect to the Redis server or Cluster.

        :rtype: tornado.concurrent.Future

        """
        LOGGER.debug('Creating a%s connection to %s:%s (db %s)',
                     ' cluster node' if self._clustering else '',
                     self._hosts[0]['host'], self._hosts[0]['port'],
                     self._hosts[0].get('db', DEFAULT_DB))
        self._connect_future = concurrent.Future()
        conn = _Connection(
            self._hosts[0]['host'], self._hosts[0]['port'],
            self._hosts[0].get('db', DEFAULT_DB),
            self._read, self._on_closed, self.io_loop,
            cluster_node=self._clustering)
        self.io_loop.add_future(conn.connect(), self._on_connected)
        return self._connect_future

    def close(self):
        """Close any open connections to Redis.

        :raises: :exc:`tredis.exceptions.ConnectionError`

        """
        if not self._connected.is_set():
            raise exceptions.ConnectionError('not connected')
        self._closing = True
        if self._clustering:
            for host in self._cluster.keys():
                self._cluster[host].close()
        elif self._connection:
            self._connection.close()

    @property
    def ready(self):
        """Indicates that the client is connected to the Redis server or
        cluster and is ready for use.

        :rtype: bool

        """
        if self._clustering:
            return (all([c.connected for c in self._cluster.values()]) and
                    len(self._cluster))
        return (self._connection and
                self._connection.connected)

    def _build_command(self, parts):
        """Build the command that will be written to Redis via the socket

        :param list parts: The list of strings for building the command
        :rtype: bytes

        """
        return self._encode_resp(parts)

    def _create_cluster_connection(self, node):
        """Create a connection to a Redis server.

        :param node: The node to connect to
        :type node: tredis.cluster.ClusterNode

        """
        LOGGER.debug('Creating a cluster connection to %s:%s',
                     node.ip, node.port)
        conn = _Connection(
            node.ip, node.port, 0, self._read, self._on_closed, self.io_loop,
            cluster_node=True, read_only='slave' in node.flags,
            slots=node.slots)
        self.io_loop.add_future(conn.connect(), self._on_connected)

    def _encode_resp(self, value):
        """Dynamically build the RESP payload based upon the list provided.

        :param mixed value: The list of command parts to encode
        :rtype: bytes

        """
        if isinstance(value, bytes):
            return b''.join([b'$', ascii(len(value)).encode('ascii'), CRLF,
                             value, CRLF])
        elif isinstance(value, str):  # pragma: nocover
            return self._encode_resp(value.encode('utf-8'))
        elif isinstance(value, int):
            return self._encode_resp(ascii(value).encode('ascii'))
        elif isinstance(value, float):
            return self._encode_resp(ascii(value).encode('ascii'))
        elif isinstance(value, list):
            output = [b'*', ascii(len(value)).encode('ascii'), CRLF]
            for item in value:
                output.append(self._encode_resp(item))
            return b''.join(output)
        else:
            raise ValueError('Unsupported type: {0}'.format(type(value)))

    @staticmethod
    def _eval_expectation(command, response, future):
        """Evaluate the response from Redis to see if it matches the expected
        response.

        :param command: The command that is being evaluated
        :type command: tredis.client.Command
        :param bytes response: The response value to check
        :param future: The future representing the execution of the command
        :type future: tornado.concurrent.Future
        :return:
        """
        if isinstance(command.expectation, int) and command.expectation > 1:
            future.set_result(response == command.expectation or response)
        else:
            future.set_result(response == command.expectation)

    def _execute(self, parts, expectation=None, format_callback=None):
        """Really execute a redis command

        :param list parts: The list of command parts
        :param mixed expectation: Optional response expectation

        :rtype: :class:`~tornado.concurrent.Future`
        :raises: :exc:`~tredis.exceptions.SubscribedError`

        """
        future = concurrent.TracebackFuture()

        try:
            command = self._build_command(parts)
        except ValueError as error:
            future.set_exception(error)
            return future

        def on_locked(_):
            if self.ready:
                if self._clustering:
                    cmd = Command(command,
                                  self._pick_cluster_host(parts),
                                  expectation, format_callback)
                else:
                    LOGGER.debug('Connection: %r', self._connection)
                    cmd = Command(command, self._connection,
                                  expectation, format_callback)
                LOGGER.debug('_execute(%r, %r, %r) on %s',
                             cmd.command, expectation, format_callback,
                             cmd.connection.name)
                cmd.connection.execute(cmd, future)
            else:
                LOGGER.critical('Lock released & not ready, aborting command')

        # Wait until the cluster is ready, letting cluster discovery through
        if not self.ready and not self._connected.is_set():
            self.io_loop.add_future(
                self._connected.wait(),
                lambda f: self.io_loop.add_future(
                    self._busy.acquire(), on_locked))
        else:
            self.io_loop.add_future(self._busy.acquire(), on_locked)

        # Release the lock when the future is complete
        self.io_loop.add_future(future, lambda r: self._busy.release())
        return future

    def _on_cluster_discovery(self, future):
        """Invoked when the Redis server has responded to the ``CLUSTER_NODES``
        command.

        :param future: The future containing the response from Redis
        :type future: tornado.concurrent.Future

        """
        LOGGER.debug('_on_cluster_discovery(%r)', future)
        common.maybe_raise_exception(future)
        nodes = future.result()
        for node in nodes:
            name = '{}:{}'.format(node.ip, node.port)
            if name in self._cluster:
                LOGGER.debug('Updating cluster connection info for %s:%s',
                             node.ip, node.port)
                self._cluster[name].set_slots(node.slots)
                self._cluster[name].set_read_only('slave' in node.flags)
            else:
                self._create_cluster_connection(node)
        self._discovery = True

    def _on_closed(self):
        """Invoked by connections when they are closed."""
        self._connected.clear()
        if not self._closing:
            if self._on_close_callback:
                self._on_close_callback()
            else:
                raise exceptions.ConnectionError('closed')

    def _on_cluster_data_moved(self, response, command, future):
        """Process the ``MOVED`` response from a Redis cluster node.

        :param bytes response: The response from the Redis server
        :param command: The command that was being executed
        :type command: tredis.client.Command
        :param future: The execution future
        :type future: tornado.concurrent.Future

        """
        LOGGER.debug('on_cluster_data_moved(%r, %r, %r)',
                     response, command, future)
        parts = response.split(' ')
        name = '{}:{}'.format(*common.split_connection_host_port(parts[2]))
        LOGGER.debug('Moved to %r', name)
        if name not in self._cluster:
            raise exceptions.ConnectionError(
                '{} is not connected'.format(name))
        self._cluster[name].execute(
            command._replace(connection=self._cluster[name]), future)

    def _on_connected(self, future):
        """Invoked when connections have been established. If the client is
        in clustering mode, it will kick of the discovery step if needed. If
        not, it will select the configured database.

        :param future: The connection future
        :type future: tornado.concurrent.Future

        """
        if future.exception():
            self._connect_future.set_exception(future.exception())
            return

        conn = future.result()
        LOGGER.debug('Connected to %s (%r, %r, %r)',
                     conn.name, self._clustering, self._discovery,
                     self._connected)
        if self._clustering:
            self._cluster[conn.name] = conn
            if not self._discovery:
                self.io_loop.add_future(self.cluster_nodes(),
                                        self._on_cluster_discovery)
            elif self.ready:
                LOGGER.debug('Cluster nodes all connected')
                if not self._connect_future.done():
                    self._connect_future.set_result(True)
                self._connected.set()
        else:
            def on_selected(sfuture):
                LOGGER.debug('Initial setup and selection processed')
                if sfuture.exception():
                    self._connect_future.set_exception(sfuture.exception())
                else:
                    self._connect_future.set_result(True)
                self._connected.set()
            select_future = concurrent.Future()
            self.io_loop.add_future(select_future, on_selected)
            self._connection = conn
            cmd = Command(self._build_command(['SELECT', str(conn.database)]),
                          self._connection, None, None)
            cmd.connection.execute(cmd, select_future)

    def _on_read_only_error(self, command, future):
        """Invoked when a Redis node returns an error indicating it's in
        read-only mode. It will use the ``INFO REPLICATION`` command to
        attempt to find the master server and failover to that, reissuing
        the command to that server.

        :param command: The command that was being executed
        :type command: tredis.client.Command
        :param future: The execution future
        :type future: tornado.concurrent.Future

        """
        failover_future = concurrent.TracebackFuture()

        def on_replication_info(_):
            common.maybe_raise_exception(failover_future)
            LOGGER.debug('Failover closing current read-only connection')
            self._closing = True
            database = self._connection.database
            self._connection.close()
            self._connected.clear()
            self._connect_future = concurrent.Future()

            info = failover_future.result()
            LOGGER.debug('Failover connecting to %s:%s',
                         info['master_host'], info['master_port'])
            self._connection = _Connection(
                info['master_host'], info['master_port'],
                database, self._read,
                self._on_closed, self.io_loop, self._clustering)

            # When the connection is re-established, re-run the command
            self.io_loop.add_future(
                self._connect_future,
                lambda f: self._connection.execute(
                    command._replace(connection=self._connection), future))

            # Use the normal connection processing flow when connecting
            self.io_loop.add_future(self._connection.connect(),
                                    self._on_connected)

        if self._clustering:
            command.connection.set_readonly(True)

        LOGGER.debug('%s is read-only, need to failover to new master',
                     command.connection.name)

        cmd = Command(self._build_command(['INFO', 'REPLICATION']),
                      self._connection,
                      None, common.format_info_response)

        self.io_loop.add_future(failover_future, on_replication_info)
        cmd.connection.execute(cmd, failover_future)

    def _read(self, command, future):
        """Invoked when a command is executed to read and parse its results.
        It will loop on the IOLoop until the response is complete and then
        set the value of the response in the execution future.

        :param command: The command that was being executed
        :type command: tredis.client.Command
        :param future: The execution future
        :type future: tornado.concurrent.Future

        """
        response = self._reader.gets()
        if response is not False:
            if isinstance(response, hiredis.ReplyError):
                if response.args[0].startswith('MOVED '):
                    self._on_cluster_data_moved(
                        response.args[0], command, future)
                elif response.args[0].startswith('READONLY '):
                    self._on_read_only_error(command, future)
                else:
                    future.set_exception(exceptions.RedisError(response))
            elif command.callback is not None:
                future.set_result(command.callback(response))
            elif command.expectation is not None:
                self._eval_expectation(command, response, future)
            else:
                future.set_result(response)
        else:
            def on_data(data):
                # LOGGER.debug('Read %r', data)
                self._reader.feed(data)
                self._read(command, future)

            command.connection.read(on_data)

    def _pick_cluster_host(self, value):
        """Selects the Redis cluster host for the specified value.

        :param mixed value: The value to use when looking for the host
        :rtype: tredis.client._Connection

        """
        crc = crc16.crc16(self._encode_resp(value[1])) % HASH_SLOTS
        for host in self._cluster.keys():
            for slot in self._cluster[host].slots:
                if slot[0] <= crc <= slot[1]:
                    return self._cluster[host]
        LOGGER.debug('Host not found for %r, returning first connection',
                     value)
        host_keys = sorted(list(self._cluster.keys()))
        return self._cluster[host_keys[0]]


class RedisClient(Client):
    """This is provided for backwards compatibility for versions < 0.7.

    .. deprecated:: 0.7

    :param str host: The hostname to connect to
    :param int port: The port to connect on
    :param int db: The database number to use
    :param method on_close: The method to call if the connection is closed
    :param bool clustering: Toggle the cluster support in the client
    :param bool auto_connect: Toggle the auto-connect on creation feature

    """
    def __init__(self,
                 host=DEFAULT_HOST,
                 port=DEFAULT_PORT,
                 db=DEFAULT_DB,
                 on_close=None,
                 clustering=False,
                 auto_connect=True):
        super(RedisClient, self).__init__(
            [{'host': host, 'port': port, 'db': db}],
            on_close, clustering=clustering, auto_connect=auto_connect)



"""Redis Server Commands Mixin"""
from tornado import concurrent

from tredis import common, exceptions

# Python 2 support for ascii()
if 'ascii' not in dir(__builtins__):  # pragma: nocover
    from tredis.compat import ascii


class ServerMixin(object):
    """Redis Server Commands Mixin"""

    def auth(self, password):
        """Request for authentication in a password-protected Redis server.
        Redis can be instructed to require a password before allowing clients
        to execute commands. This is done using the ``requirepass`` directive
        in the configuration file.

        If the password does not match, an
        :exc:`~tredis.exceptions.AuthError` exception
        will be raised.

        :param password: The password to authenticate with
        :type password: :class:`str`, :class:`bytes`
        :rtype: bool
        :raises: :exc:`~tredis.exceptions.AuthError`,
                 :exc:`~tredis.exceptions.RedisError`

        """
        future = concurrent.TracebackFuture()

        def on_response(response):
            """Process the redis response

            :param response: The future with the response
            :type response: tornado.concurrent.Future

            """
            exc = response.exception()
            if exc:
                if exc.args[0] == b'invalid password':
                    future.set_exception(exceptions.AuthError(exc))
                else:
                    future.set_exception(exc)
            else:
                future.set_result(response.result())

        execute_future = self._execute([b'AUTH', password], b'OK')
        self.io_loop.add_future(execute_future, on_response)
        return future

    def echo(self, message):
        """Returns the message that was sent to the Redis server.

        :param message: The message to echo
        :type message: :class:`str`, :class:`bytes`
        :rtype: bytes
        :raises: :exc:`~tredis.exceptions.RedisError`

        """
        return self._execute([b'ECHO', message])

    def info(self, section=None):
        """The INFO command returns information and statistics about the server
        in a format that is simple to parse by computers and easy to read by
        humans.

        The optional parameter can be used to select a specific section of
        information:

            - server: General information about the Redis server
            - clients: Client connections section
            - memory: Memory consumption related information
            - persistence: RDB and AOF related information
            - stats: General statistics
            - replication: Master/slave replication information
            - cpu: CPU consumption statistics
            - commandstats: Redis command statistics
            - cluster: Redis Cluster section
            - keyspace: Database related statistics

        It can also take the following values:

            - all: Return all sections
            - default: Return only the default set of sections

        When no parameter is provided, the default option is assumed.

        :param str section: Optional
        :return: dict

        """
        cmd = [b'INFO']
        if section:
            cmd.append(section)
        return self._execute(cmd, format_callback=common.format_info_response)

    def ping(self):
        """Returns ``PONG`` if no argument is provided, otherwise return a copy
        of the argument as a bulk. This command is often used to test if a
        connection is still alive, or to measure latency.

        If the client is subscribed to a channel or a pattern, it will instead
        return a multi-bulk with a ``pong`` in the first position and an empty
        bulk in the second position, unless an argument is provided in which
        case it returns a copy of the argument.

        :rtype: bytes
        :raises: :exc:`~tredis.exceptions.RedisError`

        """
        return self._execute([b'PING'])

    def quit(self):
        """Ask the server to close the connection. The connection is closed as
        soon as all pending replies have been written to the client.

        :rtype: bool
        :raises: :exc:`~tredis.exceptions.RedisError`

        """
        self._closing = True
        return self._execute([b'QUIT'], b'OK')

    def select(self, index=0):
        """Select the DB with having the specified zero-based numeric index.
        New connections always use DB ``0``.

        :param int index: The database to select
        :rtype: bool
        :raises: :exc:`~tredis.exceptions.RedisError`
        :raises: :exc:`~tredis.exceptions.InvalidClusterCommand`

        """
        if self._clustering:
            raise exceptions.InvalidClusterCommand
        future = self._execute(
            [b'SELECT', ascii(index).encode('ascii')], b'OK')

        def on_selected(f):
            self._connection.database = index
        self.io_loop.add_future(future, on_selected)
        return future

    def time(self):
        """Retrieve the current time from the redis server.

        :rtype: float
        :raises: :exc:`~tredis.exceptions.RedisError`

        """
        def format_response(value):
            """Format a TIME response into a datetime.datetime

            :param list value: TIME response is a list of the number
                of seconds since the epoch and the number of micros
                as two byte strings
            :rtype: float

            """
            seconds, micros = value
            return float(seconds) + (float(micros) / 1000000.0)
        return self._execute([b'TIME'], format_callback=format_response)



"""Redis HyperLogLog Commands Mixin"""


class HyperLogLogMixin(object):
    """Redis HyperLogLog Commands Mixin"""

    def pfadd(self, key, *elements):
        """Adds all the element arguments to the HyperLogLog data structure
        stored at the variable name specified as first argument.

        As a side effect of this command the HyperLogLog internals may be
        updated to reflect a different estimation of the number of unique items
        added so far (the cardinality of the set).

        If the approximated cardinality estimated by the HyperLogLog changed
        after executing the command, :meth:`~tredis.RedisClient.pfadd` returns
        ``1``, otherwise ``0`` is returned. The command automatically creates
        an empty HyperLogLog structure (that is, a Redis String of a specified
        length and with a given encoding) if the specified key does not exist.

        To call the command without elements but just the variable name is
        valid, this will result into no operation performed if the variable
        already exists, or just the creation of the data structure if the key
        does not exist (in the latter case ``1`` is returned).

        For an introduction to HyperLogLog data structure check
        :meth:`~tredis.RedisClient.pfcount`.

        .. versionadded:: 0.2.0

        .. note:: **Time complexity**: ``O(1)`` to add every element.

        :param key: The key to add the elements to
        :type key: :class:`str`, :class:`bytes`
        :param elements: One or more elements to add
        :type elements: :class:`str`, :class:`bytes`
        :rtype: bool
        :raises: :exc:`~tredis.exceptions.RedisError`

        """
        return self._execute([b'PFADD', key] + list(elements), 1)

    def pfcount(self, *keys):
        """When called with a single key, returns the approximated cardinality
        computed by the HyperLogLog data structure stored at the specified
        variable, which is ``0`` if the variable does not exist.

        When called with multiple keys, returns the approximated cardinality of
        the union of the HyperLogLogs passed, by internally merging the
        HyperLogLogs stored at the provided keys into a temporary HyperLogLog.

        The HyperLogLog data structure can be used in order to count unique
        elements in a set using just a small constant amount of memory,
        specifically 12k bytes for every HyperLogLog (plus a few bytes for the
        key itself).

        The returned cardinality of the observed set is not exact, but
        approximated with a standard error of 0.81%.

        For example in order to take the count of all the unique search queries
        performed in a day, a program needs to call
        :meth:`~tredis.RedisCount.pfcount` every time a query is processed. The
        estimated number of unique queries can be retrieved with
        :meth:`~tredis.RedisCount.pfcount` at any time.

        .. note:: as a side effect of calling this function, it is possible
           that the HyperLogLog is modified, since the last 8 bytes encode the
           latest computed cardinality for caching purposes. So
           :meth:`~tredis.RedisCount.pfcount` is technically a write command.

        .. versionadded:: 0.2.0

        .. note:: **Time complexity**: ``O(1)`` with every small average
           constant times when called with a single key. ``O(N)`` with ``N``
           being the number of keys, and much bigger constant times, when
           called with multiple keys.

        :param keys: One or more keys
        :type keys: :class:`str`, :class:`bytes`
        :rtype: int
        :returns: The approximated number of unique elements observed
        :raises: :exc:`~tredis.exceptions.RedisError`

        """
        return self._execute([b'PFCOUNT'] + list(keys))

    def pfmerge(self, dest_key, *keys):
        """Merge multiple HyperLogLog values into an unique value that will
        approximate the cardinality of the union of the observed Sets of the
        source HyperLogLog structures.

        The computed merged HyperLogLog is set to the destination variable,
        which is created if does not exist (defaulting to an empty
        HyperLogLog).

        .. versionadded:: 0.2.0

        .. note::

           **Time complexity**: ``O(N)`` to merge ``N`` HyperLogLogs, but
           with high constant times.

        :param dest_key: The destination key
        :type dest_key: :class:`str`, :class:`bytes`
        :param keys: One or more keys
        :type keys: :class:`str`, :class:`bytes`
        :rtype: bool
        :raises: :exc:`~tredis.exceptions.RedisError`

        """
        return self._execute([b'PFMERGE', dest_key] + list(keys), b'OK')
